/*
 * Copyright 2008 Perforce Software.  All rights reserved.
 *
 * This file is part of Perforce - the FAST SCM System.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

#include "MyDTG.h"
#include "MyDTS.h"
extern "C" {
#include <dtg-utils.h>
}
#include <dtg-str.h>

extern void print_DTGField( const struct DTGField *list );

const char *MyDTG::MyDTGMagic = "MyDTGClass";

struct DTGAttribute *MyDTG::list_attrs()
{

	DTGAttribute *in_field = new_DTGAttribute( 
		"redmine_user", 
		"Redmine username",
		"Set this if your Redmine user "
		"name is different from your MySQL user.", 
		NULL, 
		0 );
	in_field = append_DTGAttribute( in_field, new_DTGAttribute( 
		"rm_db",
		"MySQL database name",
		"The MySQL database name for the Redmine database, "
		"if different than \"redmine\".", 
		"redmine", 
		0 ) );
#ifdef NOT_FOR_BETA
	in_field = append_DTGAttribute( in_field, new_DTGAttribute(
		"use_repository",
		"Use repository plug-in",
		"If the Redmine project is using a 3rd party Perforce "
		"repositoy plug-in, enable this to replicate the "
		"relationships between Perforce jobs and changelists in "
		"Redmine.",
		"n",
		0 ) );
#endif
	in_field = append_DTGAttribute( in_field, new_DTGAttribute( 
		"rm_cf",
		"Fixes custom field",
		"Redmine custom Issues field name for Perforce "
		"fix information.  When in doubt, leave it be.", 
		"P4Fixes", 
		0 ) );
	in_field = append_DTGAttribute( in_field, new_DTGAttribute(
		"unicode", 
		"Unicode server",
		"Use this attribute to specify that server handles UTF-8. "
		"By default, the plug-in will use UTF-8. "
		"Specify either 'y' or 'n'.",
		"y", 
		0 ) );
	in_field = append_DTGAttribute( in_field, new_DTGAttribute(
		"offset", 
		"Change offset",
		"This attribute controls the overlap of time used when "
		"requesting a list of changed defects. If this overlap is "
		"too small, some defects will be missed by the replication. "
		"If it is too large, the replicator will process the same "
		"issue many times impacting somewhat the performance of "
		"the system. The overlap defaults to 0. Valid range is between "
		"0 and 7200 seconds. The server date returned by the Redmine "
		"server is adjusted by this offset to affect the overlap.",
		"0",  
	        0 ) );
	in_field = append_DTGAttribute( in_field, new_DTGAttribute(
		"wait_time",
		"Wait time",
		"Specifies the number of seconds that the replication engine "
		"waits after detecting a connection error before it retries "
		"an operation. Use -1 to default to the General Wait Duration "
		"specified for the replication map.",
		"10",
		0 ) );
	in_field = append_DTGAttribute( in_field, new_DTGAttribute(
		"ck_privs",
		"Check privileges",
		"Check MySQL access rights before connecting.",
		"y",
		0 ) );

	return in_field;
}

static int is_number( const char *str )
{
	if( !str )
	    return 0;

	if( *str == '.' )
	    return 0;

	// a negative number?
	if( strlen( str ) > 1 )
	{
	    if( *str == '-' && !isdigit( *(str + 1 )) )
	        return 0;
	    else
	        str += 1;
	}

	for( const char *i = str; *i; i++ )
	    if( !isdigit( *i ) )
	        return 0;

	return 1;
}

char *MyDTG::validate_attr( const struct DTGField *attr )
{
	if( !attr->name )
	    return strdup( "Unknown attribute: <>" );

	if( !strcmp( attr->name, "unicode" ) )
	    if( attr->value &&
	        ( *attr->value == 'y' || *attr->value == 'n' ) && 
	        !attr->value[1] )
	        return NULL;
	    else
	        return strdup( "Unicode server: Must be either 'y' or 'n'." );

	if( !strcmp( attr->name, "ck_privs" ) )
	    if( attr->value &&
	        ( *attr->value == 'y' || *attr->value == 'n' ) && 
	        !attr->value[1] )
	        return NULL;
	    else
	        return strdup( "Check privileges: Must be either 'y' or 'n'." );

	if( !strcmp( attr->name, "use_repository" ) )
	    if( attr->value &&
	        ( *attr->value == 'y' || *attr->value == 'n' ) && 
	        !attr->value[1] )
	        return NULL;
	    else
	        return strdup( "Check privileges: Must be either 'y' or 'n'." );
    	
	if( !strcmp( attr->name, "offset" ) )
	    if( !attr->value || !is_number( attr->value ) )
	        return strdup( "Change offset: Must be a number" );
	    else if( atoi( attr->value ) < 0 || atoi( attr->value ) > 7200 )
	        return strdup( "Change offset: Must be between 0 and 7200" );
	    else
	        return NULL;

	if( !strcmp( attr->name, "wait_time" ) )
	    if( attr->value && is_number( attr->value ) )
	        if( atoi( attr->value ) == 0 )
	            return strdup( "Wait time: Must not be zero." );
	        else if( atoi( attr->value ) < -1 )
	            return strdup( "Wait time: Must not be less than -1." );
	        else if( atoi( attr->value ) > 600 )
	            return strdup( "Wait time:  Must be less than 10 minutes.");
	        else
	            return NULL;
	    else
	        return strdup( "Wait time: Must be a positive integer,"
	                           " or -1." );

	if( !strcmp( attr->name, "redmine_user" ) ||
	    !strcmp( attr->name, "rm_cf" )     ||
	    !strcmp( attr->name, "rm_db" )
	  )
	    return NULL; // OK

	char *tmp = (char *)malloc( 20 + strlen( attr->name) );
	sprintf( tmp, "Unknown attribute: %s", attr->name );
	return tmp;
}

int MyDTG::accept_utf8()
{
	return dts->utf8_ok();
}

int MyDTG::server_offline( struct DTGError *error )
{
	return dts->server_offline( error );
}

MyDTG::MyDTG( const char *server, const char *user, const char *pass,
	    const struct DTGField *attrs,
	    struct DTGError *error  )
{
	cur_message = NULL;
	cur_message_level = 10; // None
	server_version = NULL;
	magic = MyDTGMagic;

	if( server && !strcmp( server, "*server*") &&
	    user && !strcmp( user, "*userid*") &&
	    pass && !strcmp( pass, "*passwd*") )
	{
	    testing = 1;
	    clear_DTGError( error );
	    return;
	}
	testing = 0;

	char *err = NULL;

	/* REDMINE-DTG and 1.0 are for any monitor of process support */
	/* that might be available for connections to the dts server */
	/* redmine actually does not support showing this information */
	char *my_user = cp_string( user );
	dts = new MyDTS( server, my_user, pass, attrs, "REDMINE-DTG", "1.0", err );
	delete[] my_user;

	if( err )
	{
	    set_DTGError( error, err );
	    error->can_continue = dts->is_valid();
	    delete[] err;
	}
	else
	    clear_DTGError( error );
}

MyDTG::~MyDTG()
{
	if( server_version )
	    delete[] server_version;
	if( cur_message )
	    delete[] cur_message;
	if( !testing )
	    delete dts;
}

struct DTGDate *MyDTG::extract_date( const char *date_string )
{
	/* XXX Change to support the format used by the dts */
        /*          2006-05-12 12:07:12 */
	/* example: YYYY/MM/DD HH:MM:SS */
	/*          0123456789012345678 */
	/*          2006-08-24 10:14:04 */
	int len;

	/* YYYY/MM/DD is 10 characters */
	if( !date_string || (len = strlen( date_string )) < 10 )
	    return NULL;
	struct DTGDate *date = new_DTGDate(0,0,0,0,0,0);
	date->month = atoi( &date_string[5] );
	date->day = atoi( &date_string[8] );
	date->year = atoi( &date_string[0] );

	/* YYYY/MM/DD_HH:MM is 16 characters */
	if( len > 15 )
	{
	    date->hour = atoi( &date_string[11] );
	    date->minute = atoi( &date_string[14] );
	}

	/* YYYY/MM/DD_HH:MM:SS is 19 characters */
	if( len > 18 )
	    date->second = atoi( &date_string[17] );

	return date;
}

char *MyDTG::format_date( struct DTGDate *date )
{
	if( !date )
	    return NULL;

	/* XXX Change to support the format used by the dts */
	/* return value will be deallocated using free() */
	char *date_string = (char *)malloc( sizeof(char)*32 ); 
	sprintf( date_string, "%4.4d-%2.2d-%2.2d %2.2d:%2.2d:%2.2d",
		date->year, date->month, date->day,
		date->hour, date->minute, date->second );
	return date_string;
}

struct DTGDate *MyDTG::get_server_date( struct DTGError *error )
{
	if( testing )
	{
	    clear_DTGError( error );
	    return new_DTGDate( 2006, 5, 6, 12, 34, 56 );
	}

	char *err = NULL;
	char *tmp = dts->get_server_date( err );
	if( err )
	{
	    set_DTGError( error, err );
	    error->can_continue = dts->is_valid();
	    delete[] err;
	    return NULL;
	}
	clear_DTGError( error );
	struct DTGDate *server_date = extract_date( tmp );
	delete[] tmp;

	return server_date;
}

const char *MyDTG::get_name( struct DTGError *error )
{
	clear_DTGError( error );
	return "Redmine MySQL 5"; 
}

const char *mod_version =
  "Rev. redminemysql5/"ID_OS"/"ID_REL"/"ID_PATCH" "ID_Y" "ID_M" "ID_D;
const char *MyDTG::get_module_version( struct DTGError *error )
{
	clear_DTGError( error );
	return mod_version;
}

const char *MyDTG::get_server_version( struct DTGError *error )
{
	if( testing )
	{
	    clear_DTGError( error );
	    return "RedmineTest 2.0";
	}

	char *err = NULL;
	clear_DTGError( error );
	if( server_version )
	    return server_version;

	server_version = dts->get_server_version( err );
	if( err )
	{
	    set_DTGError( error, err );
	    error->can_continue = dts->is_valid();
	    delete[] err;
	}
	return server_version;
}

char *MyDTG::get_server_warnings( struct DTGError *error )
{
	if( testing )
	{
	    clear_DTGError( error );
	    return strdup( "No warnings for RedmineTest 2.0" );
	}

	clear_DTGError( error );
	return NULL;
}

int MyDTG::get_message( struct DTGError *error )
{
	clear_DTGError( error );
	if( testing )
	{
	    set_DTGError( error, "Test message" );
	    return 0;
	}

	if( cur_message )
	{
	    set_DTGError( error, cur_message );
	    delete[] cur_message;
	    cur_message = NULL;
	    int level = cur_message_level;
	    cur_message_level = 10;
	    return level;
	}
	return 10; // No message
}

struct DTGStrList *MyDTG::list_projects( struct DTGError *error )
{
	struct DTGStrList *list;
	if( testing )
	{
	    list = new_DTGStrList( "*project*" );
	    clear_DTGError( error );
	    return list;
	}

	char *err = NULL;
	clear_DTGError( error );
	list = dts->list_projects( err );
	if( err )
	{
	    set_DTGError( error, err );
	    error->can_continue = dts->is_valid();
	    delete[] err;
	}
	return list;
}

MyDTGProj *MyDTG::get_project( const char *project, struct DTGError *error )
{
	if( !project || !*project )
	{
	    set_DTGError( error, "Undefined project" );
	    return NULL;
	}

	if( testing )
	{
	    if( strcmp( project, "*project*" ) )
	    {
	        set_DTGError( error, "Unknown project" );
	        return NULL;
	    }

	    MyDTGProj *proj = new MyDTGProj( this, project, error );
	    if( error->message )
	    {
	        delete proj;
	        proj = NULL;
	    }
	    return proj;
	}
	
	if( dts->valid_project( project ) != 0 )
	{
	    set_DTGError( error, "Unknown project");
	    return NULL;
	}

	MyDTGProj *proj = new MyDTGProj( this, project, error );
	if( error->message )
	{
	    delete proj;
	    proj = NULL;
	}
	return proj;
}
